// CPConvert.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "CPConvert.h"

extern "C" BOOL APIENTRY DllMain(
	HANDLE hModule, 
	DWORD  ul_reason_for_call, 
    LPVOID lpReserved)
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}

extern "C" CPCONVERT_API int CPConvert(int nCPFrom, int nCPTo, LPCTSTR lpszFrom, LPTSTR lpszTo, int nLenOut)
{
	_TCHAR buff[2*FORMAT_MESSAGE_MAX_WIDTH_MASK];
	DWORD dwCodeErr = 0;
	LPWSTR lpszTxtUni = NULL;
	LPSTR lpStrOut = NULL;
	int nLen = 0;
	int nLenDest = 0;
	BOOL bIsErr = FALSE;
	buff[0] = (_TCHAR)'\0';
	*(lpszTo) = (_TCHAR)'\0';

	try
	{
		// Check ValidCodePages
		if (!::IsValidCodePage((UINT)nCPFrom))
			throw _T("Error CodePage CPFrom");
		if (!::IsValidCodePage((UINT)nCPTo))
			throw _T("Error CodePage nCPTo");
		// ... end size output buffer
		if (nLenOut <= 0)
			throw _T("Error size output buffer");

		///////////////////////////////////////////////////////
		// Convert to Unicode (from code page nCPFrom)
		size_t nSizeInp = _tcslen(lpszFrom)*sizeof(_TCHAR);
		if (nSizeInp)
		{
			// Get size of WideChar buffer
			int nLenInp = 0;
			if ((nLenInp = ::MultiByteToWideChar(
					(UINT)nCPFrom,		// code page (from)
					0,					// character-type options 
					(LPCSTR)lpszFrom,	// address of string to map 
					nSizeInp,			// number of bytes in string
					0,					// address of wide-character buffer 
					0					// size of buffer 
				)) == 0) 
			{
				dwCodeErr = GetLastError();
				throw _T("Error Convert to Unicode[1]");
			}

			// Create WideChar buffer
			lpszTxtUni = ::new WCHAR[++nLenInp];
			if (lpszTxtUni == NULL)
			{
				dwCodeErr = ::GetLastError();
				throw _T("Error out of memory for unicode buffer");
			}
			
			// Convert to WideChar buffer
			if ((nLen = ::MultiByteToWideChar(
					(UINT)nCPFrom,		// code page (from)
					0,					// character-type options 
					(LPCSTR)lpszFrom,	// address of string to map 
					nSizeInp,			// number of bytes in string
					lpszTxtUni,			// address of wide-character buffer 
					nLenInp				// size of buffer 
				)) == 0) 
			{
				dwCodeErr = ::GetLastError();
				throw _T("Error Convert to Unicode[2]");
			}
			*(lpszTxtUni+nLen) = (WCHAR)'\0';
			// Ok!

			/////////////////////////////////////////////////////
			// Convert from Unicode (to code page nCPTo)
			LPCSTR lpzsDefaultChar = "?";
			BOOL bUsedDefaultChar = FALSE;
			
			// Get size of MultiByte buffer
			int nSizeOut = 0;
			if ((nSizeOut = ::WideCharToMultiByte(
					(UINT)nCPTo,		// code page (to)
					0,					// performance and mapping flags 
					lpszTxtUni,			// address of wide-character string 
					nLen,				// number of characters in string 
					NULL,				// address of buffer for new string 
					0,					// size of buffer 
					lpzsDefaultChar,	// address of default for unmappable characters 
					&bUsedDefaultChar	// address of flag set when default char. used 
 				)) == 0) 
			{
				dwCodeErr = ::GetLastError();
				throw _T("Error Convert from Unicode[1]");
			}

			// Create MultiByte buffer
			lpStrOut = ::new CHAR[++nSizeOut];
			if (lpStrOut == NULL)
			{
				dwCodeErr = ::GetLastError();
				throw _T("Error out of memory for result");
			}
			
			// Convert to MultiByte buffer
			if ((nLenDest = WideCharToMultiByte(
					(UINT)nCPTo,		// code page (to)
					0,					// performance and mapping flags 
					lpszTxtUni,			// address of wide-character string 
					nLen,				// number of characters in string 
					lpStrOut,			// address of buffer for new string 
					nSizeOut,			// size of buffer 
					lpzsDefaultChar,	// address of default for unmappable characters 
					&bUsedDefaultChar	// address of flag set when default char. used 
 				)) == 0) 
			{
				dwCodeErr = GetLastError();
				throw _T("Error Convert from Unicode[2]");
			}
			*(lpStrOut+nLenDest) = (CHAR)'\0';
			// Ok!
		}
	}
	catch (LPCTSTR lpszMsg)
	{
		bIsErr = TRUE;
		nLenDest = 0;
		if (dwCodeErr != 0)
		{
			LPVOID lpMsgBuf;
			FormatMessage( 
				FORMAT_MESSAGE_ALLOCATE_BUFFER | 
				FORMAT_MESSAGE_FROM_SYSTEM | 
				FORMAT_MESSAGE_IGNORE_INSERTS,
				NULL,
				dwCodeErr,
				MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
				(LPTSTR) &lpMsgBuf,
				0,
				NULL 
			);
			_stprintf(buff, _T("(%ld) %s : %s"),dwCodeErr, lpszMsg, lpMsgBuf);
			LocalFree(lpMsgBuf);
		}
		else
			_stprintf(buff, _T("%s"), lpszMsg);
	}
	catch(...)
	{
		bIsErr = TRUE;
		nLenDest = 0;
		_tcscpy(buff, _T("Unexpected Error"));

	}

	if (bIsErr)
	{
		// Copy error message as result
		if (_tcslen(buff) > (size_t)nLenOut)
		{
			_tcsncpy(lpszTo, (const _TCHAR*)buff, (size_t)nLenOut);
			*(lpszTo+nLenOut) = (_TCHAR)'\0';
			nLenDest = -nLenDest;
		}
		else
		{
			_tcscpy(lpszTo, (const _TCHAR*)buff);
			nLenDest = -(int)_tcslen(lpszTo);
		}
	}
	else
	{
		if (lpStrOut != NULL)
		{
			if (nLenDest != 0)
			{
				// Copy result
				// ** Because of it's not unicode-version A2TC(lpStrOut) not used **
				if (nLenDest > nLenOut)
				{
					_tcsncpy(lpszTo, lpStrOut, (size_t)nLenOut);
					*(lpszTo+nLenOut) = (_TCHAR)'\0';
				}
				else
				{
					_tcscpy(lpszTo, lpStrOut);
				}
			}
			else
			{
				*(lpszTo) = (_TCHAR)'\0';
			}
		}
	}
	if (lpszTxtUni != NULL)
	{
		::delete [] lpszTxtUni;
		lpszTxtUni = NULL;
	}

	if (lpStrOut != NULL)
	{
		::delete [] lpStrOut;
		lpStrOut = NULL;
	}
	return nLenDest;
}
